﻿using FTDI.D2XX.Device;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Diagnostics;

namespace DeviceWatcher
{
    public interface IFTEveCmd
    {
        //CMD cmd();
        uint Size();
        byte[] ToBuffer();
    }


    public enum DLprim : byte
    {
        BITMAPS = 1,    // Bitmap drawing primitive
        POINTS = 2,    // Point drawing primitive
        LINES = 3,    // Line drawing primitive
        LINE_STRIP = 4,    // Line strip drawing primitive
        EDGE_STRIP_R = 5,     // Edge strip right side drawing primitive
        EDGE_STRIP_L = 6,    // Edge strip left side drawing primitive
        EDGE_STRIP_A = 7,    // Edge strip above drawing primitive
        EDGE_STRIP_B = 8,    // Edge strip below side drawing primitive
        RECTS = 9,    // Rectangle drawing primitive
    }

    public class EveDL : MPSSEHAL
    {
        //private MPSSEHAL hal;

        // This code is provided as an example only and is not guaranteed by FTDI. FTDI accept no responsibility for any 
        // issues resulting from its use. The developer of the final application incorporating any parts of this sample project 
        // is entirely responsible for ensuring its safe and correct operation and any conequences resulting from its use.

        private const uint CMDBUF_SIZE = 4096;

        private Mutex pipeMutex;
        private Queue<IFTEveCmd> cmdPipe = new Queue<IFTEveCmd>();
       

        enum RAM : uint
        {
            CMD = 1081344,
            DL = 1048576,
            G = 0,
            PAL = 1056768,
            REG = 1057792
        }

        public enum REG : uint
        {
            ANALOG = 1058104,
            ANA_COMP = 1058160,
            BIST_CMD = 1058124,
            BIST_EN = 1058132,
            BIST_RESULT = 1058128,
            BUSYBITS = 1058008,
            CLOCK = 1057800,
            CMD_DL = 1058028,
            CMD_READ = 1058020,
            CMD_WRITE = 1058024,
            CPURESET = 1057820,
            CRC = 1058152,
            CSPREAD = 1057892,
            CYA0 = 1058000,
            CYA1 = 1058004,
            CYA_TOUCH = 1058100,
            DATESTAMP = 1058108,
            DITHER = 1057884,
            DLSWAP = 1057872,
            FRAMES = 1057796,
            FREQUENCY = 1057804,
            GPIO = 1057936,
            GPIO_DIR = 1057932,
            HCYCLE = 1057832,
            HOFFSET = 1057836,
            HSIZE = 1057840,
            HSYNC0 = 1057844,
            HSYNC1 = 1057848,
            ID = 1057792,
            INT_EN = 1057948,
            INT_FLAGS = 1057944,
            INT_MASK = 1057952,
            MACRO_0 = 1057992,
            MACRO_1 = 1057996,
            MARCH_ACC = 1058144,
            MARCH_DIR = 1058136,
            MARCH_OP = 1058140,
            MARCH_WIDTH = 1058148,
            OUTBITS = 1057880,
            PCLK = 1057900,
            PCLK_POL = 1057896,
            PLAY = 1057928,
            PLAYBACK_FORMAT = 1057972,
            PLAYBACK_FREQ = 1057968,
            PLAYBACK_LENGTH = 1057960,
            PLAYBACK_LOOP = 1057976,
            PLAYBACK_PLAY = 1057980,
            PLAYBACK_READPTR = 1057964,
            PLAYBACK_START = 1057956,
            PWM_DUTY = 1057988,
            PWM_HZ = 1057984,
            RENDERMODE = 1057808,
            ROMSUB_SEL = 1058016,
            ROTATE = 1057876,
            SNAPSHOT = 1057816,
            SNAPY = 1057812,
            SOUND = 1057924,
            SWIZZLE = 1057888,
            TAG = 1057912,
            TAG_X = 1057904,
            TAG_Y = 1057908,
            TAP_CRC = 1057824,
            TAP_MASK = 1057828,
            TOUCH_ADC_MODE = 1058036,
            TOUCH_CHARGE = 1058040,
            TOUCH_DIRECT_XY = 1058164,
            TOUCH_DIRECT_Z1Z2 = 105816,
            TOUCH_MODE = 1058032,
            TOUCH_OVERSAMPLE = 1058048,
            TOUCH_RAW_XY = 1058056,
            TOUCH_RZ = 1058060,
            TOUCH_RZTHRESH = 1058052,
            TOUCH_SCREEN_XY = 1058064,
            TOUCH_SETTLE = 1058044,
            TOUCH_TAG = 1058072,
            TOUCH_TAG_XY = 1058068,
            TOUCH_TRANSFORM_A = 1058076,
            TOUCH_TRANSFORM_B = 1058080,
            TOUCH_TRANSFORM_C = 1058084,
            TOUCH_TRANSFORM_D = 1058088,
            TOUCH_TRANSFORM_E = 1058092,
            TOUCH_TRANSFORM_F = 1058096,
            TRACKER = 1085440,
            TRIM = 1058156,
            VCYCLE = 1057852,
            VOFFSET = 1057856,
            VOL_PB = 1057916,
            VOL_SOUND = 1057920,
            VSIZE = 1057860,
            VSYNC0 = 1057864,
            VSYNC1 = 1057868
        }



        public enum DL : byte
        {
            BEGIN = 0x1f,         // Begin drawing a graphics primitive
            BITMAP_HANDLE = 0x05,         // Specify the bitmap handle
            BITMAP_LAYOUT = 0x07,         // Specify the source bitmap memory format and layout for the current handle
            BITMAP_SIZE = 0x08,         // Specify the screen drawing of bitmaps for the current handle
            BITMAP_SOURCE = 0x01,         // Specify the source address of bitmap data
            BITMAP_TRANSFORM_A = 0x15,
            BITMAP_TRANSFORM_B = 0x16,
            BITMAP_TRANSFORM_C = 0x17,
            BITMAP_TRANSFORM_D = 0x18,
            BITMAP_TRANSFORM_E = 0x19,
            BITMAP_TRANSFORM_F = 0x1A,
            BLEND_FUNC = 0x0B,
            CALL = 0x1D,
            CELL = 0x06,
            CLEAR = 0x26,
            CLEAR_COLOR_A = 0x0F,
            CLEAR_COLOR_RGB = 0x02,
            COLOR_RGB = 0x04,

            CLEAR_STENCIL = 0x11,
            CLEAR_TAG = 0x12,
            
            // ...

            DISPLAY = 0x00,
            END = 0x21,
            LINE_WIDTH = 0x0E,
            POINT_SIZE = 0x0D,
        }



        const byte FT_GPU_INTERNAL_OSC = 0x48; //default
        const byte FT_GPU_EXTERNAL_OSC = 0x44;



        const byte FT_GPU_PLL_48M = 0x62;  //default
        const byte FT_GPU_PLL_36M = 0x61;
        const byte FT_GPU_PLL_24M = 0x64;



        const byte FT_GPU_ACTIVE_M = 0x00;
        const byte FT_GPU_STANDBY_M = 0x41;//default
        const byte FT_GPU_SLEEP_M = 0x42;
        const byte FT_GPU_POWERDOWN_M = 0x50;


        const byte FT_GPU_CORE_RESET = 0x68;





        public EveDL()
        {
            pipeMutex = new Mutex(false, "pipeMutex");
        }

        async new public Task Configure(FTD2xxDevice myDevice)
        {
            await base.Configure(myDevice);
            await InitEve();
        }

        // ============================================================================================================
        // Send address of a register which is to be Written
        // ============================================================================================================

        // This function sends an address of a register which is to be written. It sets the MSBs to 10 to indicate a write  
        //
        // Pass in:         00000000 xxxxxxxx yyyyyyyy zzzzzzzz     address as double word 32-bit value, only lower 3 bytes used
        //
        // Sends:           10xxxxxx                                bits 23 - 16 of the address, with bit 23/22 forced to 10
        //                  yyyyyyyy                                bits 15 - 8  of the address
        //                  zzzzzzzz                                bits 7 -  0  of the address
        //
        // Returns:         N/A


        async public Task SendAddressWR(uint memoryAddress)

        {
            byte SPI_Writebyte = 0x00;
            byte a, b, c;
           
            SPI_Writebyte = (byte)((memoryAddress & 0x00FF0000) >> 16);    // Mask off the first byte to send
            a= (byte)(SPI_Writebyte & 0xBF | 0x80);           // Since this is a write, the MSBs are forced to 10
            
            b = (byte)((memoryAddress & 0x0000FF00) >> 8);     // Mask off the next byte to be sent
            
            c = (byte)(memoryAddress & 0x000000FF);            // Mask off the next byte to be sent (least significant byte)
            await ReadWrite(new byte[] { a, b, c });                                   // Call the low-level SPI routine for this MCU to send this byte                         

        }

        // ============================================================================================================
        // Send address of a register which is to be Read
        // ============================================================================================================

        // This function sends an address of a register which is to be read. It clears the two MS bits to indicate a read
        //
        // Pass in:         00000000 xxxxxxxx yyyyyyyy zzzzzzzz     address as double word 32-bit value, only lower 3 bytes used
        //
        // Sends:           00xxxxxx                                bits 23 - 16 of the address, with bits 23/22 forced to 0
        //                  yyyyyyyy                                bits 15 - 8  of the address
        //                  zzzzzzzz                                bits 7 -  0  of the address
        //                  00000000                                dummy byte 0x00 written as described in FT800 datasheet
        //
        // Returns:         No return value
        async public Task SendAddressRD(uint memoryAddress)
        {
            byte SPI_Writebyte = 0x00;
            byte a, b, c;
            // Write out the address. Only the lower 3 bytes are sent, with the most significant byte sent first
            SPI_Writebyte = (byte)((memoryAddress & 0x00FF0000) >> 16);     // Mask off the first byte to send
            a = (byte)(SPI_Writebyte & 0x3F);                   // Since this is a read, the upper two bits are forced to 00
            b= (byte)((memoryAddress & 0x0000FF00) >> 8);      // Mask off the next byte to be sent
            c = (byte)(memoryAddress & 0x000000FF);             // Mask off the next byte to be sent (least significant byte)
            //await ReadWrite(new byte[] { a, b, c });                                   // Call the low-level SPI routine for this MCU to send this byte           
            await ReadWrite(new byte[] { a, b, c, 0x00 });                                   // Call the low-level SPI routine for this MCU to send this byte           
        }


        // ============================================================================================================
        // Read an 8-bit register
        // ============================================================================================================

        // This function reads the actual 8-bit data from a register. Call FT800_SPI_SendAddressRD first to specify the address
        // 
        // Pass in:         N/A
        //
        // Sends:           00000000                                Dummy write 0x00 whilst clocking in the byte 
        //
        // Returns:         aaaaaaaa                                Byte which was clocked in
        async public Task<byte> ReadByte()
        {
            byte[] SPI_Writebyte = { 0x00 };
            byte[] SPI_Readbyte = new byte[1];
            // Read the data byte. We pass a 0x00 into the SPI routine since it always writes when it reads
            //SPI_Writebyte = 0x00;                                   // Write a dummy 0x00
            SPI_Readbyte = await ReadWrite(SPI_Writebyte);        // Get the byte which was read by the low-level MCU routine            
            return (SPI_Readbyte[0]);// Return the byte which we read
        }

        async public Task<uint> ReadUint()
        {
            byte[] SPI_Writebytes = new byte[4] { 0x00, 0x00, 0x00, 0x00 };
            byte[] SPI_Readbytes = new byte[4];
            uint SPI_DWordRead = 0x00000000;
            uint DwordTemp = 0x00000000;
            SPI_Readbytes = await ReadWrite(SPI_Writebytes);            // Get the byte which was read by the low-level MCU routine
            DwordTemp = (uint)((SPI_Readbytes[3] << 24) | (SPI_Readbytes[2] << 16) | (SPI_Readbytes[1] << 8) | SPI_Readbytes[0]);                               // Put received byte into a temporary 32-bit value
            SPI_DWordRead = DwordTemp;                     // Put the byte into a 32-bit variable (shifted 24 bits up)         
            return (SPI_DWordRead);   // Return the word

        }
        async public Task<ushort> ReadWord()
        {
            byte[] SPI_Writebytes = new byte[2] { 0x00, 0x00 };
            byte[] SPI_Readbytes = new byte[2];
            ushort SPI_DWordRead = 0x00000000;

            SPI_Readbytes = await ReadWrite(SPI_Writebytes);            // Get the byte which was read by the low-level MCU routine
            SPI_DWordRead = (ushort)((uint)(SPI_Readbytes[1] << 8) | (uint)(SPI_Readbytes[0]));  // Put received byte into a temporary 32-bit value                    
            return (SPI_DWordRead);   // Return the word
        }

        // write a byte array
        //
        //
        //
        /// <summary>
        /// 
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        async public Task Write(byte[] value)
        {
            await ReadWrite(value);                       // Write the last (most significant) byte
        }
        // ============================================================================================================
        // Write a 32-bit value
        // ============================================================================================================

        // This function writes a 32-bit value. Call FT800_SPI_SendAddressWR first to specify the register address
        //
        // Pass in:         pppppppp qqqqqqqq rrrrrrrr ssssssss             32-bit value to be written to the register
        //
        // Sends:           ssssssss         least signicant byte first
        //                  rrrrrrrr        
        //                  qqqqqqqq        
        //                  pppppppp         most significant byte last
        //  
        // Returns:         N/A
        async public Task Write(uint value)
        {
            byte[] SPI_Writebytes = new byte[] { 0x00, 0x00, 0x00, 0x00 };

            SPI_Writebytes[0] = (byte)(value & 0x000000FF);              //
            SPI_Writebytes[1] = (byte)((value & 0x0000FF00) >> 8);       //
            SPI_Writebytes[2] = (byte)((value & 0x00FF0000) >> 16);                     //
            SPI_Writebytes[3] = (byte)((value & 0xFF000000) >> 24);      //
            await ReadWrite(SPI_Writebytes);                       // Write the last (most significant) byte
        }

        // ============================================================================================================
        // Write a 16-bit value
        // ============================================================================================================

        // This function writes a 16-bit value. Call FT800_SPI_SendAddressWR first to specify the register address
        //
        // Pass in:         rrrrrrrr ssssssss               16-bit value to be written to the register
        //
        // Sends:           ssssssss                        least significant byte first
        //                  rrrrrrrr                        most significant byte last
        //  
        // Returns:         N/A

        async public Task Write(ushort value)
        {
            byte[] SPI_Writebytes = new byte[] { 0x00, 0x00 };

            SPI_Writebytes[0] = (byte)(value & 0x000000FF);              //
            SPI_Writebytes[1] = (byte)((value & 0x0000FF00) >> 8);       //          
            await ReadWrite(SPI_Writebytes);                       // Write the last (most significant) byte
        }

        // ============================================================================================================
        // Write an 8-bit value
        // ============================================================================================================

        // This function writes an 8-bit value. Call FT800_SPI_SendAddressWR first to specify the register address
        //
        // Pass in:         ssssssss          16-bit value to be written to the register
        //
        // Sends:           ssssssss          sends the specified byte
        //  
        // Returns:         N/A
        async public Task Write(byte value)
        {
            byte[] SPI_Writebytes = new byte[] { 0x00 };

            SPI_Writebytes[0] = (byte)(value & 0x000000FF);              //
            await ReadWrite(SPI_Writebytes);                       // Write the last (most significant) byte
        }



        // ============================================================================================================
        // Write a Host Command
        // ============================================================================================================

        // This function is specifically designed to send a Host Command
        //
        // Pass in:         xxhhhhhh             8-bit host command where upper 2 bits are dont-care
        //
        // Sends:           [Chip Select Low]
        //                  01hhhhhh         Host command with bit 7 forced to 0 and bit 6 forced to 1
        //                  00000000         Dummy 0x00        
        //                  00000000         Dummy 0x00 
        //                  [Chip Select High] 
        //  
        // Returns:         N/A

        private async Task HostCommand(byte Host_Command)
        {
            byte[] readbyte = { 0x00 };
            byte[] writeByte = { 0x00 };

            // Chip Select Low
            await CSLow();

            writeByte[0] = (byte)(Host_Command & 0x3F | 0x40);   // This is the command being sent
            await ReadWrite(writeByte);                       //

            writeByte[0] = 0x00;                                 // Sending dummy byte
            await ReadWrite(writeByte);                       //             
            await ReadWrite(writeByte);                       //

            // Chip Select High
            await CSHigh();
        }


        // ============================================================================================================
        // Host Command Dummy Read
        // ============================================================================================================

        // This function is specifically designed to send three 0x00 bytes to wake up the FT800
        //
        // Pass in:         N/A             
        //
        // Sends:           [Chip Select Low]
        //                  00000000         Dummy 0x00 
        //                  00000000         Dummy 0x00        
        //                  00000000         Dummy 0x00 
        //                  [Chip Select High] 
        //  
        // Returns:         N/A

        async private Task HostCommandDummyRead()
        {
            byte[] byteRead = { 0x00 };
            byte[] writeByte = { 0x00 };

            // Chip Select Low
            await CSLow();

            // Read/Write           
            byteRead = await ReadWrite(writeByte);
            byteRead = await ReadWrite(writeByte);
            byteRead = await ReadWrite(writeByte);

            // Chip Select High
            await CSHigh();
        }

        async public Task InitEve()
        {

            Debug.WriteLine("Init Eve+");
            //await PDHigh();
            //await CSHigh();
            //await Task.Delay(3);

            await PDLow();
            await Task.Delay(20);                                        // Delay for power up of regulator
            await PDHigh();
            await Task.Delay(20);                                        // Delay for power up of regulator

            await HostCommandDummyRead();               // Read location 0 to wake up FT800

            await HostCommand(FT_GPU_EXTERNAL_OSC);     // Change the PLL to external clock - optional

            await HostCommand(FT_GPU_PLL_48M);          //  Ensure configured to 48 MHz

            // Note: Could now increase SPI clock rate up to 30MHz SPI if preferred

            await Task.Delay(20);                                        // Delay 

            await HostCommand(FT_GPU_CORE_RESET);       // Reset the core

            await Task.Delay(20);                                        // Delay 

            await HostCommand(FT_GPU_ACTIVE_M);         // Read address 0 to ensure FT800 is active

            byte chipid = 0x00;                         // Read the Chip ID to check comms with the FT800 - should be 0x7C
            while (chipid != 0x7C)
            {
                await CSLow();                      // CS low
                await SendAddressRD((uint)REG.ID);
                await Task.Delay(100);
                // Send the address
                chipid = await ReadByte();              // Read the actual value
                await CSHigh();                     // CS high
            }

            // =======================================================================
            // Write the display registers on the FT800 for your particular display
            // =======================================================================

            //await CSLow();                                // CS low                   Write REG_HCYCLE to 548
            //await SendAddressWR((uint)REG.HCYCLE);            // Send the address
            //await Write((ushort)548);                         // Send the 16-bit value
            //await CSHigh();                                     // CS high
            await  MemWrite((uint)REG.HCYCLE, (ushort)548);                         // Send the 16-bit value
                                                                                    //Write REG_HCYCLE to 548
            //await CSLow();                                // CS low                   Write REG_HOFFSET to 43
            //await SendAddressWR((uint)REG.HOFFSET);           // Send the address
            //await Write((ushort)43);                          // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.HOFFSET,(ushort)43);


            //await CSLow();                                // CS low                   Write REG_HSYNC0 to 0
            //await SendAddressWR((uint)REG.HSYNC0);            // Send the address
            //await Write((ushort)0);                           // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.HSYNC0, (ushort)0);    // Write REG_HSYNC0 to 0

            //await CSLow();                                // CS low                   Write REG_HSYNC1 to 41
            //await SendAddressWR((uint)REG.HSYNC1);            // Send the address
            //await Write((ushort)41);                          // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.HSYNC1, (ushort)41);  //Write REG_HSYNC1 to 41

            //await CSLow();                                // CS low                   Write REG_VCYCLE to 292
            //await SendAddressWR((uint)REG.VCYCLE);            // Send the address
            //await Write((ushort)292);                         // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.VCYCLE, (ushort)292);  // Write REG_VCYCLE to 292

            //await CSLow();                                // CS low                   Write REG_VOFFSET to 12
            //await SendAddressWR((uint)REG.VOFFSET);           // Send the address
            //await Write((ushort)12);                          // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.VOFFSET, (ushort)12);  //Write REG_VOFFSET to 12

            //await CSLow();                                // CS low                   Write REG_VSYNC0 to 0
            //await SendAddressWR((uint)REG.VSYNC0);            // Send the address
            //await Write((ushort)0);                           // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.VSYNC0, (ushort)0);    //Write REG_VSYNC0 to 0

            //await CSLow();                                // CS low                   Write REG_VSYNC1 to 10
            //await SendAddressWR((uint)REG.VSYNC1);            // Send the address
            //await Write((ushort)10);                          // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.VSYNC1, (ushort)10 );  //Write REG_VSYNC1 to 10


            //await CSLow();                                // CS low                   Write REG_SWIZZLE to 2
            //await SendAddressWR((uint)REG.SWIZZLE);           // Send the address
            //await Write((ushort)2);                           // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.SWIZZLE, (ushort)2 );  // Write REG_SWIZZLE to 2


            //await CSLow();                                // CS low                   Write REG_PCLK_POL to 1
            //await SendAddressWR((uint)REG.PCLK_POL);          // Send the address
            //await Write((ushort)1);                           // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.PCLK_POL, (ushort)1 ); // Write REG_PCLK_POL to 1

            //await CSLow();                                // CS low                   Write REG_HSIZE to 480
            //await SendAddressWR((uint)REG.HSIZE);             // Send the address
            //await Write((ushort)480);                         // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite( (uint)REG.HSIZE, (ushort)480 );   //Write REG_HSIZE to 480

            //await CSLow();                                // CS low                   Write REG_VSIZE to 272
            //await SendAddressWR((uint)REG.VSIZE);             // Send the address
            //await Write((ushort)272);                         // Send the 16-bit value
            //await CSHigh();                               // CS high
            await MemWrite((uint)REG.VSIZE, (ushort)272);   //Write REG_VSIZE to 272




            //await CSLow();                                // CS low                   Write REG_PCLK to 5
            //await SendAddressWR((uint)REG.PCLK);              // Send the address
            //await Write((ushort)5);                           // Send the 16-bit value
            //await CSHigh();
            await MemWrite( (uint)REG.PCLK, (ushort)5); // Write REG_PCLK to 5

            // =======================================================================
            // Send an initial display list which will clear the screen to a black colour
            // ======================================================================= 



            // Set the colour which is used when the colour buffer is cleared        
            //await CSLow();                                //
            //await SendAddressWR((uint)RAM.DL + 0);            // Write first entry in Display List (first location in DL Ram)
            //await Write(0x02808080);                  // Clear Color RGB   00000010 BBBBBBBB GGGGGGGG RRRRRRRR  (B/G/R = Colour values)
            //await CSHigh();                               //
            await MemWrite( (uint)RAM.DL + 0, 0x02808080);                  // Clear Color RGB   00000010 BBBBBBBB GGGGGGGG RRRRRRRR  (B/G/R = Colour values)

            // Clear the Colour, Stencil and Tag buffers. This will set the screen to the 'clear' colour set above.
            //await CSLow();                                //
            //await SendAddressWR((uint)RAM.DL + 4);            // Write next entry in display list (each command is 4 bytes)
            //await Write(0x26000007);                  // Clear 00100110 -------- -------- -----CST  (C/S/T define which parameters to clear)
            //await CSHigh();                               //
            await MemWrite( (uint)RAM.DL + 4, 0x26000007);                  // Clear 00100110 -------- -------- -----CST  (C/S/T define which parameters to clear)

            // Display command ends the display list
            //await CSLow();                                //
            //await SendAddressWR((uint)RAM.DL + 8);            // Write next entry in display list 
            //await Write(0x00000000);                  // DISPLAY command 00000000 00000000 00000000 00000000
            //await CSHigh();                               //
            await MemWrite((uint)RAM.DL + 8, 0x00000000);  // DISPLAY command 00000000 00000000 00000000 00000000

            // Writing to the DL_SWAP register tells the Display Engine to render the new screen designed above.
            //await CSLow();                                //         
            //await SendAddressWR((uint)REG.DLSWAP);            // Writing to the DL_SWAP register...value 10 means render after last frame complete        
            //await Write(0x00000002);                  // 00000000 00000000 00000000 000000SS  (SS bits define when render occurs)
            //await CSHigh();                               // CS high
            await MemWrite( (uint)REG.DLSWAP, 0x00000002);                  // 00000000 00000000 00000000 000000SS  (SS bits define when render occurs)




            // =======================================================================
            // Set the GPIO pins of the FT800 to enable the display now
            // =======================================================================

            // note: Refer to the GPIO section in the FT800 datasheet and also the connections to the GPIO
            // used on the particular development module being used

            // Bit 7 enables the LCD Display. It is set to output driving high
            // Bit 1 enables the Audio Amplifier. It is set to output driving low to shut down the amplifier since audio not used here
            // Bit 0 is unused but set to output driving high.

            //await CSLow();                              // CS low                   Write REG_GPIO_DIR
            //await SendAddressWR((uint)REG.GPIO_DIR);        // Send the address
            //await Write(0x83);                              // Send the 8-bit value
            //await CSHigh();                             // CS high
            await MemWrite( (uint)REG.GPIO_DIR, 0x83); 

            //await CSLow();                              // CS low                   Write REG_GPIO
            //await SendAddressWR((uint)REG.GPIO);            // Send the address
            //await Write(0x81);                              // Send the 8-bit value
            //await CSHigh();                             // CS high
            await MemWrite((uint)REG.GPIO, 0x81);
            StartCommandProc();
            Debug.WriteLine("Init Eve-!");
        }


        // ============================================================================================================
        // Increment Command Offset
        // ============================================================================================================

        // This function is used when adding commands to the Command FIFO.
        // After adding a command to the FIFO, this function will calculate the FIFO address offset where the next command
        // should be written to.
        // In general, the next command would be written to [previous offset plus the length of the last command] but due 
        // to the circular FIFO used, a command list may need to wrap round from an offset of 4095 to 0. 
        //
        // Pass in:         N/A             
        //
        // Sends:           [Chip Select Low]
        //                  00000000         Dummy 0x00 
        //                  00000000         Dummy 0x00        
        //                  00000000         Dummy 0x00 
        //                  [Chip Select High] 
        //  
        // Returns:         N/A


        private uint IncCMDOffset(uint currentOffset, byte commandSize)
        {
            uint newOffset;

            newOffset = currentOffset + commandSize;

            if (newOffset > 4095)
            {
                newOffset = (newOffset - 4096);
            }

            return newOffset;
        }


        async
        private Task  MemWrite (uint addr, uint data )
        {
            await CSLow();                                //
            await SendAddressWR( addr );  // Writing to next available location in FIFO (FIFO base address + offset)  
            await Write(data);
            await CSHigh();                               // 
        }

        async
        private Task MemWrite(uint addr, ushort data)
        {
            await CSLow();                                //
            await SendAddressWR( addr );  // Writing to next available location in FIFO (FIFO base address + offset)  
            await Write(data);
            await CSHigh();                               // 
        }

        async
        private Task MemWrite(uint addr, byte data)
        {
            await CSLow();                                //
            await SendAddressWR( addr );  // Writing to next available location in FIFO (FIFO base address + offset)  
            await Write(data);
            await CSHigh();                               // 
        }

        async
        private Task<uint> MemUnitRead(uint addr)
        {
            uint rval;
            await CSLow();                            
            await SendAddressRD(addr);      
            rval = await ReadUint();           
            await CSHigh();
            return rval;
        }

        async 
        private Task<ushort> MemWordRead (uint addr)
        {
            ushort rval;
            await CSLow();                            
            await SendAddressRD(addr);      
            rval = await ReadWord();        
            await CSHigh();
            return rval;
        }

        async
        private Task<byte> MemByteRead(uint addr)
        {
            byte rval;
            await CSLow();                  
            await SendAddressRD(addr);      
            rval = await ReadByte();        
            await CSHigh();
            return rval;
        }

       
        public void PushCmd(IFTEveCmd cmd)
        {
            //Lock Pipe
            pipeMutex.WaitOne();
            cmdPipe.Enqueue(cmd);
            pipeMutex.ReleaseMutex();
            //unlock pipe
        }

        private IFTEveCmd PopCmd()
        {
            IFTEveCmd cmd;
            //Lock Pipe
            pipeMutex.WaitOne();
            cmd = cmdPipe.Dequeue();
            pipeMutex.ReleaseMutex();
            //unlock pipe
            return cmd;
        }

        /// DL Class instructions
        //
        // 
        public class Begin : FTDLCmd
        {
            private DL c;
            private DLprim p;
            private uint size;


            public Begin(DLprim dlp)
            {
                c = DL.BEGIN;
                size = 4;
                p = dlp;
            }
            public DL Cmd()
            {
                return c;
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[4];
                buff[3] = (byte)c;
                buff[2] = (byte)0;
                buff[1] = (byte)0;
                buff[0] = (byte)p;
                return buff;
            }
        }
        public class Display : FTDLCmd
        {
            private DL c;
            private uint size;
            public Display()
            {
                c = DL.DISPLAY;
                size = 4;
            }

            public DL Cmd()
            {
                return c;
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)c;
                buff[2] = (byte)0;
                buff[1] = (byte)0;
                buff[0] = (byte)0;
                return buff;
            }
        }

        public class Vertex2II : FTDLCmd
        {
            private DL c;
            private uint size;
            ushort x, y;
            byte h, cl;
            public Vertex2II(ushort xcor, ushort ycor, byte handle, byte cell)
            {
                c = (DL)0x80;           // 0b10xxxxxx
                size = 4;
                x = (ushort)(xcor & 0x01FF);
                y = (ushort)(ycor & 0x01FF);
                h = handle;
                cl = cell;
            }

            public DL Cmd()
            {
                return (DL)((uint)c & 0xC0);
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)((ushort)c | ((x & 0x01F8) >> 3));              // x & 0x 1 1111 1000 >> 3
                buff[2] = (byte)(((x & 0x0007) << 5) | ((y & 0x01f0) >> 3));    // x & 0x 0 0000 0111 << 5 | y  & 0x 1 1111 0000 >>3 
                buff[1] = (byte)((y & 0x0f)<<4 | (h & 0x1C) >> 1);              //                           y  & 0x 0 0000 1111 <<4 | h & 0x0 0001 1110 >> 1
                buff[0] = (byte)((h & 0x01) << 7 | (cl & 0x3F));                // h & 0x 0000 0001   << 7 | cl & 0x 0111 1111 
                return buff;
            }

        }

        public class Vertex2F
        {
            private DL c;
            private uint size;
            ushort x, y;
            public Vertex2F(ushort xcor, ushort ycor)
            {
                c = (DL)0x40;           // 0b01xxxxxx
                size = 4;
                x = xcor;
                y = ycor;
            }

            public DL Cmd()
            {
                return (DL)((uint)c & 0xC0);
            }

             public uint Size()
            {
                return size;
            }

            public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)((x & 0x03FC) >> 2);
                buff[2] = (byte)(((x & 0x03) << 6) | ((y & 0x3F00) >> 8) * 0x3F);
                buff[1] = (byte)(y & 0x00FF);
                buff[0] = (byte)((ushort)c | (x & 0xFC) >> 2);
                return buff;
            }
        }
        public class End
        {
            private DL c;
            private uint size;
            public End()
            {
                c = DL.END;
                size = 4;
            }
            public DL Cmd()
            {
                return c;
            }
            public uint Size()
            {
                return size;
            }
            public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)c;
                buff[2] = (byte)0;
                buff[1] = (byte)0;
                buff[0] = (byte)0;
                return buff;
            }
        }
        public class LineWidth : FTDLCmd
        {
            private DL c;
            private uint size;
            private ushort lineWidth;
            public LineWidth(ushort w)
            {
                c = DL.LINE_WIDTH;
                size = 4;
                lineWidth = w;
            }
            public DL Cmd()
            {
                return c;
            }
            override public uint Size()
            {
                return size;
            }
            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)c;
                buff[2] = (byte)0;
                buff[1] = (byte)((lineWidth & 0x7F00) >> 8);
                buff[0] = (byte)(lineWidth & 0xFF);
                return buff;
            }
        }
        public class PointSize : FTDLCmd
        {
            private DL c;
            private uint size;
            private ushort pointSize;
            public PointSize(ushort w)
            {
                c = DL.POINT_SIZE;
                size = 4;
                pointSize = w;
            }
            public DL Cmd()
            {
                return c;
            }
            override public uint Size()
            {
                return size;
            }
            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)c;
                buff[2] = (byte)0;
                buff[1] = (byte)((pointSize & 0xfF00) >> 8);
                buff[0] = (byte)(pointSize & 0xFF);
                return buff;
            }
        }



        ////CLEAR_COLOR_RGB
        public class ClearColourRGB : FTDLCmd
        {
            private DL c;
            private uint size;
            private byte r, g, b;
            public ClearColourRGB(byte red, byte green, byte blue)
            {
            c = DL.CLEAR_COLOR_RGB;
                size = 4;
                r = red;
                g = green;
                b = blue;
            }
            public DL Cmd()
            {
                return c;
            }
            override public uint Size()
            {
                return size;
            }
            override  public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)c;
                buff[2] = b;
                buff[1] = g;
                buff[0] = r;        
                return buff;
            }
        }


        ////CLEAR_COLOR_RGB
        public class ColourRGB : FTDLCmd
        {
            private DL c;
            private uint size;
            private byte r, g, b;
            public ColourRGB(byte red, byte green, byte blue)
            {
                c = DL.COLOR_RGB;
                size = 4;
                r = red;
                g = green;
                b = blue;
            }
            public DL Cmd()
            {
                return c;
            }
            override public uint Size()
            {
                return size;
            }
            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)c;
                buff[2] = b;
                buff[1] = g;
                buff[0] = r;
                return buff;
            }
        }

        ////CLEAR
        public class Clear : FTDLCmd
        {
            private DL c;
            private uint size;

            public Clear()
            {
                c = DL.CLEAR;
                size = 4;
            }


            // Not Implemented !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            public Clear(byte cf,byte sf,byte tf)
            {
                c = DL.CLEAR;
                size = 4;
                throw new  NotImplementedException();
           
            }
            public DL Cmd()
            {
                return c;
            }
            override public uint Size()
            {
                return size;
            }
            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];
                buff[3] = (byte)c;
                buff[2] = 0;
                buff[1] = 0;
                buff[0] = 7;
                return buff;
            }
        }


        public abstract class FTDLCmd : IFTEveCmd
        {
            virtual public byte[] CMDToBuffer(CMD c)
            {
                byte[] buff = { (byte)( ((uint)c & 0xff000000) >> 24),
                                (byte)( ((uint)c & 0x00FF0000) >> 16),
                                (byte)( ((uint)c & 0x0000FF00) >> 8 ),
                                (byte)( ((uint)c & 0x000000FF)      )  };
                return buff;


            }

            public abstract uint Size();
            public abstract byte[] ToBuffer();
        }


        private void  StartCommandProc()
        {

            var t = Task.Factory.StartNew(async () =>
            {
                await CommandProc();
                return; 
            });
            //t.Wait();
           
        }

        private enum CPS
        {
            Init,
            Idle,
            DoCoPro,
            DoDL,
            Error
        }

        async private Task CommandProc()
        {
            CPS currentState = CPS.Init;

            try
            {
                while (true)
                {
                    switch (currentState)
                    {
                        case CPS.Init:
                            currentState = await Init(currentState);
                            break;
                        case CPS.Idle:
                            currentState = await Idle(currentState);
                            break;
                        case CPS.DoCoPro:
                            currentState = await DoCoPro(currentState);
                            break;
                        case CPS.DoDL: 
                            currentState = await DoDL(currentState);
                            break;
                        case CPS.Error:
                            currentState = await Error(currentState);
                            break;
                        default:
                            currentState = await Idle(currentState);
                            break;
                    }


                }

            }
            catch ( Exception ex)
            {
                Debug.WriteLine("CommandProc ex: {0}", ex.Message);
            }

            return;

        }

        async private Task<CPS> Init( CPS current)
        {
            return CPS.Idle;
        }
        async  private Task<CPS> Idle( CPS current )
        {
            CPS retVal = CPS.Idle;
            int count;

            pipeMutex.WaitOne();
            count = cmdPipe.Count();
            pipeMutex.ReleaseMutex();

            if (count  > 2)
            {

                pipeMutex.WaitOne();
                if ( cmdPipe.Peek() is CoPro.Dlstart)
                {
                    foreach( IFTEveCmd c in cmdPipe)
                    {
                        if (c is CoPro.Flush)
                        {
                            retVal = CPS.DoCoPro;
                            break;
                        }
                        else
                        {
                            //Debug.WriteLine("Waiting for Flush");
                            retVal = CPS.Idle;
                        }
                    }
                }
                pipeMutex.ReleaseMutex();
            }
            else
            {
                await Task.Delay(20);
                retVal = CPS.Idle;
            }

            return retVal;
        }
        
        async private Task<CPS> DoCoPro ( CPS current)
        {
            uint cmdOffset;
            uint cmdBufferWr;
            //uint cmdBufferWr2;
            uint cmdBufferRd;
            //uint cmdBufferRd2;
            IFTEveCmd cmd;
            int count;

            pipeMutex.WaitOne();
                count = cmdPipe.Count();
            pipeMutex.ReleaseMutex();

            
            if (count > 0)
            {
                // ================================================================================================================
                // Generate the Co-Processor commands for the FIRST screen
                // ================================================================================================================ 

                // Wait for the Graphics Engine to become idle by reading the REG_CMD_WRITE and REG_CMD_READ buffers 
                // REG_CMD_WRITE will show the address in the FIFO of the last command entered
                // REG_CMD_READ shows the address of the command at which the Graphics Engine has reached
                // When equal, this means that the Graphics Engine has finished executing the last command list

                // Since the FIFO is circular and the READ pointer equals the WRITE pointer, we can assume we have the full 
                // 4096 bytes of the FIFO to write this next command list.

                do
                {

                    cmdBufferWr = await MemWordRead((uint)REG.CMD_WRITE);
                    cmdBufferRd = await MemWordRead((uint)REG.CMD_READ);
                } while (cmdBufferWr != cmdBufferRd);

                while (cmdPipe.Count > 0)
                {
                  

                    byte[] tempBuffer;



                    // Check the actual value of the current WRITE register (pointer)

                    cmdOffset = cmdBufferWr;                       // Get the current value of the CMD_WRITE pointer in the FIFO


                    // ************************** This is the actual list of commands to draw the screen ***********************        

                 

                    uint msgSize=0;
                    // Accumulate Commands into a single buffer
                    pipeMutex.WaitOne();
                    foreach ( IFTEveCmd c in cmdPipe)
                    {
                        msgSize += c.Size(); 
                        if ( c is CoPro.Flush )                 
                            break;                        
                    }
                   
                    pipeMutex.ReleaseMutex();
                    tempBuffer = new byte[msgSize];
                    uint i = 0;
                    //foreach (IFTEveCmd c in cmdPipe)
                    do
                    {
                        cmd = PopCmd();
                        cmd.ToBuffer().CopyTo(tempBuffer, (int)i);
                        i += cmd.Size();
                    } while (!(cmd is CoPro.Flush));

                    // Next, define the background colour as black and clear the display parameters. This will produce a blank black screen. 
                    await CSLow();                                //
                    await SendAddressWR((uint)RAM.CMD + cmdOffset);  // Writing to next available location in FIFO (FIFO base address + offset)     
                    await Write(tempBuffer);                  //write the command to the buffer
                    await CSHigh();                               //


                    cmdOffset = IncCMDOffset(cmdOffset, (byte)msgSize); // Move the CMD Offset since we have just added 4 bytes to the FIFO                 
                                                                        //Debug.WriteLine("Blue,{0}", c++);

                    // ************************** This is the end of the list of commands to draw the screen ***********************        

                    // Currently, CMD_WRITE still equals CMD_READ.  The FT800 has still not executed any of our new commands.
                    // We now write the CMD_WRITE register to tell the FT800 where the end of our new command list is
                    // The FT800 will now work it's way through our command list until CMD_READ equals our new CMD_WRITE

                    // await CSLow();                                //
                    // await SendAddressWR((uint)REG.CMD_WRITE);         // 
                    // await Write((ushort)cmdOffset);                  //    
                    // await CSHigh();                               //


                    await MemWrite((uint)REG.CMD_WRITE, (ushort)cmdOffset);         // 





                    cmdBufferWr = await MemWordRead((uint)REG.CMD_WRITE);
                    cmdBufferRd = await MemWordRead((uint)REG.CMD_READ);
                    //Debug.WriteLine("Wr: 0x{0:x}, Rd:0x{0:x}", cmdBufferWr, cmdBufferRd);
                                                                

                }
            }
            return CPS.Idle;
        }





        async private Task<CPS> DoDL    ( CPS current)
        {
            return CPS.Idle;
        }
        async private Task<CPS> Error(CPS currentState)
        {
            return CPS.Error;
        }

    }
}





